/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * License); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * AS IS BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

/*
 * Copyright (c) 2020, OPEN AI LAB
 * Author: bzhang@openailab.com
 */

#include <math.h>
#include "sys_port.h"
#include "module.h"
#include "tengine_errno.h"
#include "tengine_log.h"
#include "tengine_ir.h"
#include "../../cpu_node_ops.h"
#include "tengine_op.h"


int ref_l2normalization_fp32(float* input_data, float* output_data, int size, int channel_size)
{
    for(int i = 0; i < size; i++)
    {
        float sq_l2_norm = 0;
        for(int j = 0; j < channel_size; j++)
        {
            const float val = input_data[j];
            sq_l2_norm += val * val;
        }
        const float l2_norm = sqrt(sq_l2_norm);
        for(int j = 0; j < channel_size; j++)
        {
            *output_data = *input_data / l2_norm;
            output_data++;
            input_data++;
        }
    }
    return 0;
}

static int init_node(struct node_ops* node_ops, struct exec_node* exec_node, struct exec_graph* exec_graph)
{
    return 0;
}

static int release_node(struct node_ops* node_ops, struct exec_node* exec_node, struct exec_graph* exec_graph)
{
    return 0;
}

static int run(struct node_ops* node_ops, struct exec_node* exec_node, struct exec_graph* exec_graph)
{
    struct ir_node* ir_node = exec_node->ir_node;
    struct ir_graph* ir_graph = ir_node->graph;
    struct ir_tensor* input_tensor;
    struct ir_tensor* output_tensor;

    input_tensor = get_ir_graph_tensor(ir_graph, ir_node->input_tensors[0]);
    output_tensor = get_ir_graph_tensor(ir_graph, ir_node->output_tensors[0]);

    int input_size = 1;
    int channel_size = input_tensor->dims[input_tensor->dim_num-1];

    for(int i = 0; i < input_tensor->dim_num; i++){
        input_size *= input_tensor->dims[i];
    }

    ref_l2normalization_fp32(input_tensor->data, output_tensor->data, input_size, channel_size);

    return 0;
}

static int reshape(struct node_ops* node_ops, struct exec_node* exec_node, struct exec_graph* exec_graph)
{
    struct ir_node* node = exec_node->ir_node;
    struct ir_graph* ir_graph = node->graph;
    struct ir_tensor* input = get_ir_graph_tensor(ir_graph, node->input_tensors[0]);
    struct ir_tensor* output = get_ir_graph_tensor(ir_graph, node->output_tensors[0]);

    int ret = set_ir_tensor_shape(output, input->dims, input->dim_num);
    return ret;
}

static int score(struct node_ops* node_ops, struct exec_graph* exec_graph, struct ir_node* exec_node)
{
    return OPS_SCORE_CANDO;
}

static struct node_ops hcl_node_ops = {.prerun = NULL,
                                       .run = run,
                                       .reshape = reshape,
                                       .postrun = NULL,
                                       .init_node = init_node,
                                       .release_node = release_node,
                                       .score = score};

static int reg_l2normalization_hcl_ops(void* arg)
{
    return register_builtin_node_ops(OP_L2NORMALIZATION, &hcl_node_ops);
}

static int unreg_l2normalization_hcl_ops(void* arg)
{
    return unregister_builtin_node_ops(OP_L2NORMALIZATION, &hcl_node_ops);
}

AUTO_REGISTER_OPS(reg_l2normalization_hcl_ops);
AUTO_UNREGISTER_OPS(unreg_l2normalization_hcl_ops);
